// #![allow(dead_code)]

mod cpu;
mod error;
mod instructions;
mod keyboard;
mod memory;
mod tui;
mod types;
mod video;

use crate::cpu::Cpu;
use crate::keyboard::Keyboard;
use crate::memory::Mem;
use crate::video::{Screen, TerminalRenderer};

use cpu::{CpuController, CpuReceiver};
use crossterm::cursor::Hide;
use crossterm::execute;
use crossterm::terminal::{size, Clear, ClearType, SetSize};
// use cpu::CpuController;
use memory::MemHandle;
// use clap::Parser;
// use minifb::{Scale, ScaleMode, Window, WindowOptions};
use serde::{Deserialize, Serialize};

use std::{
    fs::File,
    io::{stdout, Read, Result},
    path::PathBuf,
    str::FromStr,
    sync::mpsc,
    thread,
};

use crossterm::{
    cursor,
    event::{self, KeyCode, KeyEventKind},
    terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
    // ExecutableCommand,
};

// use toml::Table;

//#[derive(Parser)]
//struct Cli {
//    // Load a rom onto the system rom
//    #[arg(short, required = true)]
//    rom: Option<std::path::PathBuf>,
//}

#[derive(Serialize, Deserialize, Debug)]
struct Config {
    char_rom: Option<String>,
    rom: String,
}

fn main() -> Result<()> {
    let mut stdout = stdout();
    let (cols, rows) = size()?;
    execute!(stdout, SetSize(64, 29), cursor::Hide, EnterAlternateScreen)?;
    enable_raw_mode()?;

    let config: Config = match File::open("./config.toml") {
        Ok(mut file) => {
            let mut string = String::new();
            file.read_to_string(&mut string).unwrap();
            toml::from_str(string.as_str()).unwrap()
        }
        Err(_) => return Ok(()),
    };

    let mut memory = Mem::new();
    let rom = match std::fs::File::open(config.rom) {
        Ok(file) => file,
        Err(error) => panic!("Couldn't open main rom! {:?}", error),
    };
    if let Err(error) = memory.load_rom(rom) {
        println!("{:?}", error);
    };

    memory
        .dump(PathBuf::from_str("./coredump.bin").unwrap())
        .unwrap();

    let shared_memory = MemHandle::new(memory);
    let screen_memory = shared_memory.clone();
    let cpu_memory = shared_memory.clone();
    let keyboard_memory = shared_memory.clone();

    let (cpu_tx, cpu_rx) = mpsc::channel();
    // let (state_tx, state_rx) = mpsc::channel();
    let screen_cpu_tx = cpu_tx.clone();
    // let (window_tx, window_rx) = mpsc::channel();

    // thread::spawn(move || {
    //     let mut screen = Crtc::new(
    //         screen_memory,
    //         config.char_rom.as_ref(),
    //         CpuController::new(screen_cpu_tx),
    //         window_tx,
    //     );
    //     screen.run();
    // });

    // let mut window = Window::new(
    //     "ʕ·ᴥ·ʔ-☆",
    //     512,
    //     380,
    //     WindowOptions {
    //         resize: true,
    //         borderless: true,
    //         title: true,
    //         transparency: false,
    //         scale: Scale::X2,
    //         scale_mode: ScaleMode::AspectRatioStretch,
    //         topmost: false,
    //         none: true,
    //     },
    // )
    // .unwrap();

    // window.set_input_callback(Box::new(Keyboard::new(keyboard_memory)));

    let mut cpu = Cpu::new(cpu_memory, CpuReceiver::new(cpu_rx));
    // let mut cpu = Cpu::new(cpu_memory);
    // let mut tui = tui::App::new(
    //     CpuController::new(cpu_tx.clone()),
    //     shared_memory.clone(),
    //     state_rx,
    // );
    thread::spawn(move || {
        cpu.reset().unwrap();
        cpu.memory.write(0x4400, 0b0000_0100);
        cpu.execute()
    });

    let stdout_lock = stdout.lock();
    let renderer = TerminalRenderer::new(screen_memory, stdout_lock);
    let mut screen = Screen::new(CpuController::new(cpu_tx.clone()), renderer);

    // thread::spawn(move || {
    //     screen.run();
    // });

    loop {
        if event::poll(std::time::Duration::from_millis(16))? {
            if let event::Event::Key(key) = event::read()? {
                if key.kind == KeyEventKind::Press && key.code == KeyCode::Char('q') {
                    break;
                }
            }
        }
        // let buffer = window_rx.recv().unwrap();
        screen.draw()
        // tui.update(&mut terminal)?;
    }

    execute!(
        stdout,
        LeaveAlternateScreen,
        SetSize(cols, rows),
        Clear(ClearType::Purge)
    )?;
    disable_raw_mode()?;
    Ok(())
}
