use crate::types::{Byte, Word};
use anyhow::{bail, Result};
use std::io::{self, Write};
use std::path::PathBuf;
use std::str::FromStr;
use std::sync::{Arc, Mutex, MutexGuard};
use std::u16;
use std::{fs::File, io::Read};

#[derive(Clone, Debug)]
pub struct MemHandle(Arc<Mutex<Mem>>);
impl MemHandle {
    pub fn new(memory: Mem) -> Self {
        Self(Arc::new(Mutex::new(memory)))
    }
    pub fn read(&self, address: Word) -> Byte {
        let memory = self.lock().unwrap();
        memory.read(address)
    }
    pub fn write(&self, address: Word, data: Byte) {
        let mut memory = self.lock().unwrap();
        memory.write(address, data);
    }

    pub fn dump(&self) {
        let memory = self.lock().unwrap();
        memory.dump(PathBuf::from_str("./cpu_dump.bin").unwrap());
    }

    fn lock(&self) -> Result<MutexGuard<'_, Mem>> {
        match self.0.lock() {
            Ok(result) => Ok(result),
            Err(error) => bail!("{error}"),
        }
    }
}

// TODO: impl Iterator for MemHandle so we can get references to the underlying data from other threads without cloning

pub trait MemoryReader {
    fn read(&self, address: u16) -> u8;
}

pub trait MemoryWriter {
    fn write(&self, address: u16, data: u8);
}

// pub trait MemoryAccess: MemoryReader + MemoryWriter {}
// impl<T> MemoryAccess for T where T: MemoryReader + MemoryWriter {}

#[derive(Debug, Clone)]
pub struct Mem {
    pub data: Vec<u8>,
}

impl Default for Mem {
    fn default() -> Self {
        Self {
            data: vec![0; u16::MAX as usize + 1],
        }
    }
}

impl Mem {
    pub fn new() -> Self {
        Mem::default()
    }
    pub fn dump(&self, path: PathBuf) -> io::Result<()> {
        let mut outfile = File::create(path)?;
        outfile.write_all(&self.data)?;
        Ok(())
    }

    pub fn read(&self, address: Word) -> Byte {
        self.data[address as usize]
    }

    pub fn write(&mut self, address: Word, data: Byte) {
        self.data[address as usize] = data;
    }

    pub fn load_rom(&mut self, rom: File) -> Result<()> {
        let bytes = rom.bytes();
        for (address, byte) in bytes.enumerate() {
            match byte {
                Ok(value) => self.write(address as Word + 0x8000, value),
                Err(_) => {
                    bail!("Loading rom: couldn't write byte {:#04x}", address);
                }
            }
        }
        Ok(())
    }
    //pub fn read_from_bin(&mut self, f: File) -> Result<()> {
    //    let bytes = f.bytes();
    //    for (address, byte) in bytes.enumerate() {
    //        match byte {
    //            Ok(value) => self.write(address as Word, value),
    //            Err(_) => {
    //                bail!("couldn't write byte {:#04x}", address);
    //            }
    //        }
    //    }
    //    Ok(())
    //}
}
